#include "tunes.h"

void
lookup (hashv_t name)
     /* Lookup variable ``name'' in environment ``env'', and store its
      * value in register ``val''. */
{
 redo:
  if ( core[env].type == O_ENVNULL )
    error ("Unbound variable.");
  else if ( core[env].type != O_ENV )
    error ("INTERNAL ERROR: Environment expected!");
  else if ( core[core[env].d.env_v.def].type != O_DEF )
    error ("INTERNAL ERROR: Frame expected!");
  else if ( core[core[env].d.env_v.def].d.def_v.name == name )
    val = core[core[env].d.env_v.def].d.def_v.val;
  else
    {
      env = core[env].d.env_v.seq;
      goto redo;
    }
}

void
eval (void)
     /* Evaluate ``expr'' in environment ``env''.  Either pass the
      * result to ``cont'' (via invoke()) or directly set ``task'' to
      * what must be run (via step()) to continue the computation. */
{
  if ( core[expr].type == O_VAR )
    {
      lookup (core[expr].d.var_v);
      invoke ();
    }
  else if ( core[expr].type == O_APPLY )
    {
      task = allocate ();
      core[task].type = O_TASK_APPLY_0;
      core[task].d.task_apply_0_v.erator = core[expr].d.apply_v.erator;
      core[task].d.task_apply_0_v.erand = core[expr].d.apply_v.erand;
      core[task].d.task_apply_0_v.env = env;
      core[task].d.task_apply_0_v.cont = cont;
    }
  else if ( core[expr].type == O_PROGN )
    {
      task = allocate ();
      core[task].type = O_TASK_PROGN_0;
      core[task].d.task_progn_0_v.primus = core[expr].d.progn_v.primus;
      core[task].d.task_progn_0_v.pares = core[expr].d.progn_v.pares;
      core[task].d.task_progn_0_v.env = env;
      core[task].d.task_progn_0_v.cont = cont;
    }
  else if ( core[expr].type == O_QUOTE )
    {
      val = core[expr].d.quote_v;
      invoke ();
    }
  else if ( core[expr].type == O_LAMBDA )
    {
      val = allocate ();
      core[val].type = O_USERFUNC;
      core[val].d.userfunc_v.param = core[expr].d.lambda_v.param;
      core[val].d.userfunc_v.env = env;
      core[val].d.userfunc_v.body = core[expr].d.lambda_v.body;
      invoke ();
    }
  else
    /* Deprecated: please use QUOTE instead. */
    {
      val = expr;
      invoke ();
    }
}

void
step (void)
     /* Advance ``task''. */
{
  if ( core[task].type == O_TASK_EVAL )
    {
      expr = core[task].d.task_eval_v.expr;
      env = core[task].d.task_eval_v.env;
      cont = core[task].d.task_eval_v.cont;
      eval ();
    }
  else if ( core[task].type == O_TASK_APPLY )
    {
      erator = core[task].d.task_apply_v.erator;
      erand = core[task].d.task_apply_v.erand;
      cont = core[task].d.task_apply_v.cont;
      apply ();
    }
  else if ( core[task].type == O_TASK_PROGN_0 )
    {
      cont = allocate ();
      core[cont].type = O_CONT_EVAL;
      core[cont].d.cont_eval_v.expr
	= core[task].d.task_progn_0_v.pares;
      core[cont].d.cont_eval_v.env
	= core[task].d.task_progn_0_v.env;
      core[cont].d.cont_eval_v.cont
	= core[task].d.task_progn_0_v.cont;
      expr = core[task].d.task_progn_0_v.primus;
      env = core[task].d.task_progn_0_v.env;
      eval ();
    }
  else if ( core[task].type == O_TASK_APPLY_0 )
    {
      cont = allocate ();
      core[cont].type = O_CONT_APPLY_1;
      core[cont].d.cont_apply_1_v.erand
	= core[task].d.task_apply_0_v.erand;
      core[cont].d.cont_apply_1_v.env
	= core[task].d.task_apply_0_v.env;
      core[cont].d.cont_apply_1_v.cont
	= core[task].d.task_apply_0_v.cont;
      expr = core[task].d.task_apply_0_v.erator;
      env = core[task].d.task_apply_0_v.env;
      eval ();
    }
  else if ( core[task].type == O_TASK_APPLY_1 )
    {
      cont = allocate ();
      core[cont].type = O_CONT_APPLY;
      core[cont].d.cont_apply_v.erator
	= core[task].d.task_apply_1_v.erator;
      core[cont].d.cont_apply_v.cont
	= core[task].d.task_apply_1_v.cont;
      expr = core[task].d.task_apply_1_v.erand;
      env = core[task].d.task_apply_1_v.env;
      eval ();
    }
  else
    error ("INTERNAL ERROR: Trying to run a non-task.");
}

void
invoke (void)
     /* Invoke a ``cont'': set ``task'' to do the work. */
{
  if ( core[cont].type == O_CONT_FINISH )
    no_error ("Execution terminated.");
  else if ( core[cont].type == O_CONT_EVAL )
    {
      task = allocate ();
      core[task].type = O_TASK_EVAL;
      core[task].d.task_eval_v.expr = core[cont].d.cont_eval_v.expr;
      core[task].d.task_eval_v.env = core[cont].d.cont_eval_v.env;
      core[task].d.task_eval_v.cont = core[cont].d.cont_eval_v.cont;
    }
  else if ( core[cont].type == O_CONT_APPLY )
    {
      task = allocate ();
      core[task].type = O_TASK_APPLY;
      core[task].d.task_apply_v.erator = core[cont].d.cont_apply_v.erator;
      core[task].d.task_apply_v.erand = val;
      core[task].d.task_apply_v.cont = core[cont].d.cont_apply_v.cont;
    }
  else if ( core[cont].type == O_CONT_APPLY_1 )
    {
      task = allocate ();
      core[task].type = O_TASK_APPLY_1;
      core[task].d.task_apply_1_v.erator = val;
      core[task].d.task_apply_1_v.erand
	= core[cont].d.cont_apply_1_v.erand;
      core[task].d.task_apply_1_v.env
	= core[cont].d.cont_apply_1_v.env;
      core[task].d.task_apply_1_v.cont
	= core[cont].d.cont_apply_1_v.cont;
    }
  else
    error ("INTERNAL ERROR: Tried to invoke a non-continuation.");
}

void
extend_env (hashv_t name)
{
  temp1 = allocate ();
  core[temp1].type = O_DEF;
  core[temp1].d.def_v.name = name;
  core[temp1].d.def_v.val = val;
  temp2 = allocate ();
  core[temp2].type = O_ENV;
  core[temp2].d.env_v.def = temp1;
  core[temp2].d.env_v.seq = env;
  env = temp2;
  temp1 = a_unit ();
  temp2 = a_unit ();
}

void
apply (void)
{
  if ( core[erator].type == O_PRIMFUNC )
    apply_primfunc (core[erator].d.primfunc_v.nm);
  else if ( core[erator].type == O_USERFUNC )
    {
      expr = core[erator].d.userfunc_v.body;
      val = erand;
      env = core[erator].d.userfunc_v.env;
      extend_env (core[erator].d.userfunc_v.param);
      eval ();
    }
  else if ( core[erator].type >= O_CONT_FINISH
	    && core[erator].type < O_TASK_EVAL )
    {
      cont = erator;
      val = erand; /* Calling continuation: discard cont */
      invoke ();
    }
  else
    error ("Wrong type to apply.");
}
