#include <stdio.h> /* Debugging only */
#include "tunes.h"

#define CORESIZE 1048576
#define HCORESIZE 524288

struct object_s core[CORESIZE];

static objptr_t freeptr = 0;

void do_gc (void);

char may_gc = 0;

objptr_t
allocate (void)
     /* Return a newly allocated object.  If may_gc is set (normal
      * case) this may trigger garbage-collection.  WARNING: this
      * means that all your local variables (and those of the
      * functions that called you) will become invalid. */
{
  objptr_t obj;

  if ( freeptr == HCORESIZE || freeptr == CORESIZE )
    {
      if ( ! may_gc )
	error ("Need to GC and not permitted.");
      do_gc ();
    }
  obj = freeptr++;
  return obj;
}

objptr_t
gc_relocate (objptr_t obj)
     /* Relocate a useful object during garbage collection, and
      * replace its old location by a broken heart pointing to its new
      * location. */
{
  if ( core[obj].type == O_BROKEN_HEART )
    /* Already relocated */
    return core[obj].d.broken_heart_v;
  core[freeptr] = core[obj];
  core[obj].type = O_BROKEN_HEART;
  core[obj].d.broken_heart_v = freeptr;
  return freeptr++;
}

void
do_gc (void)
     /* Perform garbage collection: relocate all global variables, and
      * then scan core for useful data.  This function is quite long
      * because we have so many different object types.  Conceptually,
      * it is very simple. */
{
  objptr_t scan0;
  objptr_t scan;

  fprintf (stderr, "Starting garbage collection... ");
  if ( freeptr == CORESIZE )
    freeptr = 0;
  scan = scan0 = freeptr;
  gunit = gc_relocate (gunit);
  mainprog = gc_relocate (mainprog);
  globenv = gc_relocate (globenv);
  curtask = gc_relocate (curtask);
  final = gc_relocate (final);
  task = gc_relocate (task);
  expr = gc_relocate (expr);
  env = gc_relocate (env);
  cont = gc_relocate (cont);
  erator = gc_relocate (erator);
  erand = gc_relocate (erand);
  val = gc_relocate (val);
  temp1 = gc_relocate (temp1);
  temp2 = gc_relocate (temp2);
  temp3 = gc_relocate (temp3);
  temp4 = gc_relocate (temp4);
  for ( ; scan<freeptr ; scan++ )
    switch ( core[scan].type )
      {
      case O_PAIR:
	core[scan].d.pair_v.car = gc_relocate (core[scan].d.pair_v.car);
	core[scan].d.pair_v.cdr = gc_relocate (core[scan].d.pair_v.cdr);
	break;
      case O_REF:
	core[scan].d.ref_v = gc_relocate (core[scan].d.ref_v);
	break;
      case O_DEF:
	core[scan].d.def_v.val = gc_relocate (core[scan].d.def_v.val);
	break;
      case O_ENV:
	core[scan].d.env_v.def = gc_relocate (core[scan].d.env_v.def);
	core[scan].d.env_v.seq = gc_relocate (core[scan].d.env_v.seq);
	break;
      case O_USERFUNC:
	core[scan].d.userfunc_v.env = gc_relocate (core[scan].d.userfunc_v.env);
	core[scan].d.userfunc_v.body = gc_relocate (core[scan].d.userfunc_v.body);
	break;
      case O_CONT_APPLY:
	core[scan].d._2val.d1 = gc_relocate (core[scan].d._2val.d1);
	core[scan].d._2val.d2 = gc_relocate (core[scan].d._2val.d2);
	break;
      case O_CONT_EVAL:
      case O_CONT_APPLY_1:
      case O_TASK_EVAL:
      case O_TASK_APPLY:
	core[scan].d._3val.d1 = gc_relocate (core[scan].d._3val.d1);
	core[scan].d._3val.d2 = gc_relocate (core[scan].d._3val.d2);
	core[scan].d._3val.d3 = gc_relocate (core[scan].d._3val.d3);
	break;
      case O_TASK_APPLY_0:
      case O_TASK_APPLY_1:
      case O_TASK_PROGN_0:
	core[scan].d._4val.d1 = gc_relocate (core[scan].d._4val.d1);
	core[scan].d._4val.d2 = gc_relocate (core[scan].d._4val.d2);
	core[scan].d._4val.d3 = gc_relocate (core[scan].d._4val.d3);
	core[scan].d._4val.d4 = gc_relocate (core[scan].d._4val.d4);
	break;
      case O_APPLY:
	core[scan].d.apply_v.erator = gc_relocate (core[scan].d.apply_v.erator);
	core[scan].d.apply_v.erand = gc_relocate (core[scan].d.apply_v.erand);
	break;
      case O_PROGN:
	core[scan].d.progn_v.primus = gc_relocate (core[scan].d.progn_v.primus);
	core[scan].d.progn_v.pares = gc_relocate (core[scan].d.progn_v.pares);
	break;
      case O_QUOTE:
	core[scan].d.quote_v = gc_relocate (core[scan].d.quote_v);
	break;
      case O_LAMBDA:
	core[scan].d.lambda_v.body = gc_relocate (core[scan].d.lambda_v.body);
	break;
      default:
	/* Nothing to do */
	break;
      }
  fprintf (stderr, "%ld objects of non-garbage found.\n", scan-scan0);
}

static char gunit_defined = 0;

objptr_t
a_unit (void)
     /* Return (and intialize if needed) the gunit variable (should
      * always point to a unit object). */
{
  if ( ! gunit_defined )
    {
      objptr_t obj;

      obj = allocate ();
      core[obj].type = O_UNIT;
      return gunit = obj;
    }
  else
    return gunit;
}
